"""This module implements methods for defining functions
which map vectors to vectors (of same or different lengths).

Examples include simple transformations such as translation or rotation,
and complicated mappings such as the creation of Preference Incorporated Spaces.
"""

__all__ = ["classificationPIS"]

from desdeo_tools.maps.preference_incorporated_space_RP import classificationPIS
